# @title new perm cca

import numpy as np
from permcca.inference import permutation_inference
from sklearn.decomposition import PCA
from sklearn.preprocessing import StandardScaler
from sklearn.feature_selection import VarianceThreshold
from sklearn.pipeline import Pipeline
from sklearn.cross_decomposition import CCA, PLSCanonical, PLSRegression
from scipy.io import loadmat

# X = loadmat("../k10_r3_X.mat")["data"]
# Y = loadmat("../k10_r3_Y.mat")["data"]
X = np.load("../assets/X.npy")
Y = np.load("../assets/Y.npy")

x_preproc = Pipeline(
    [
        ("scale", StandardScaler(with_mean=True, with_std=True)),
        ("var", VarianceThreshold(threshold=0.0)),
        # ("pca", PCA(n_components=5)),
    ]
).fit_transform(X)

y_preproc = Pipeline(
    [
        ("scale", StandardScaler(with_mean=True, with_std=True)),
        ("var", VarianceThreshold(threshold=0.0)),
        # ("pca", PCA(n_components=5)),
    ]
).fit_transform(Y)

# U_hat, V_hat = permutation_inference(x_preproc, y_preproc)
# p = permutation_inference(x_preproc, y_preproc)
# print(p)
from permcca.preproc import get_max_ndim

n_comps = get_max_ndim(x_preproc, y_preproc)
ps = []
for model in [
    CCA(n_components=n_comps),
    PLSCanonical(n_components=n_comps),
    PLSRegression(n_components=n_comps),
]:
    ps.append(permutation_inference(x_preproc, y_preproc, model=model))
#%%
from sklearn.cross_decomposition import CCA
from scipy.linalg import null_space

X = x_preproc.copy()
Y = y_preproc.copy()
n_comps = min(np.linalg.matrix_rank(X), np.linalg.matrix_rank(Y))

cca = CCA(n_components=n_comps, scale=False)
cca.fit_transform(X, Y)
U_hat = X @ np.hstack((cca.x_weights_, null_space(cca.x_weights_.T)))
V_hat = Y @ np.hstack((cca.y_weights_, null_space(cca.y_weights_.T)))



import numpy as np
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.utils.validation import check_array

"""
The way of permutation in CCA analysis is shuffling the identity matrix.
This can be simplified to just shuffling the ids (or indexes) of one feature set, it is the same.
"""


class Permutation(BaseEstimator, TransformerMixin):
    # Shuffling the ids of one feature set

    def __init__(self, random_state=42, shuffle_target=False):
        if isinstance(random_state, (int, np.integer)):
            self.random_state = random_state
        else:
            raise ValueError("random_state must be an integer or None.")
        self.shuffle_target = shuffle_target
        self.perm_ids_x = None
        self.perm_ids_y = None

    def permute(self, ids_len):
        rng = np.random.default_rng(self.random_state)
        self.perm_ids_x = rng.permutation(ids_len)
        if self.shuffle_target:
            self.perm_ids_y = rng.permutation(ids_len)

    def fit(self, x: np.ndarray, y: np.ndarray or None = None):
        self.permute(x.shape[0])
        return self

    def transform(self, x, y=None):
        if y is not None:
            check_array(y)
            if self.shuffle_target:
                return x[self.perm_ids_x, :], y[self.perm_ids_y, :]
            return x[self.perm_ids_x, :], y
        return x[self.perm_ids_x, :]

    def fit_transform(self, X, y=None, **fit_params):
        return self.fit(X, y).transform(X, y)


class PermutationMultiVar(Permutation):
    def __init__(self, random_state=42, shuffle_target=False, selection=None):
        super().__init__(random_state=random_state, shuffle_target=shuffle_target)
        if isinstance(selection, (list, np.ndarray)) or selection is None:
            self.selection = selection
        else:
            raise ValueError("selection must be a list, np.ndarray or None.")

    def _select_rows(self, x):
        # avoid the structured dependency already exists in the data, for example, the family cohort
        if self.selection is None:
            return x
        elif isinstance(self.selection, list):
            # check dimension
            if x.shape[0] != len(self.selection):
                raise ValueError(
                    "The number of rows in x must be equal to the length of selection."
                )
            else:
                return x[self.selection, :]
        elif isinstance(self.selection, np.ndarray):
            # if the selection is a 1d array, it is the same as a list
            if self.selection.ndim == 1:
                if x.shape[0] != len(self.selection):
                    raise ValueError(
                        "The number of rows in x must be equal to the length of selection."
                    )
                else:
                    return x[self.selection, :]
            elif self.selection.ndim == 2:
                if x.shape[0] != self.selection.shape[1]:
                    raise ValueError(
                        "The number of cols in x must be equal to the number of cols in selection."
                    )
                else:
                    return self.selection @ x

    def fit(self, x: np.ndarray, y: np.ndarray = None):
        selected = self._select_rows(x)
        self.permute(selected.shape[0])
        return self

    def transform(self, x, y=None):
        selected_x = self._select_rows(x)
        if y is not None:
            check_array(y)
            selected_y = self._select_rows(y)
            if self.shuffle_target:
                return selected_x[self.perm_ids_x, :], selected_y[self.perm_ids_y, :]
            return selected_x[self.perm_ids_x, :], selected_y
        return selected_x[self.perm_ids_x, :]



import numpy as np
from sklearn.cross_decomposition import CCA
from scipy.linalg import null_space
from scipy.stats import pearsonr
from permcca.permute import PermutationMultiVar
from typing import Callable, Union, Tuple, Dict, Any, Optional
from sklearn.base import BaseEstimator
import pingouin as pg


def can_corr(X_c: np.ndarray, Y_c: np.ndarray):
    return np.array([pearsonr(X_c[:, i], Y_c[:, i])[0] for i in range(X_c.shape[1])])


def _wilk_statistic(can_corr: np.ndarray):
    return -np.flip(np.cumsum(np.flip(np.log(1 - can_corr**2))))[0]


def _calc_cum_rs(
    U_hat: np.ndarray,
    V_hat: np.ndarray,
    n_comps: int,
    model: BaseEstimator,
):
    latent_cum_rs = []
    for k in range(n_comps):
        try:
            working_model = model.set_params(
                **{"n_components": n_comps - k}
            )  # sklearn models
        except AttributeError:
            raise NotImplementedError("Currently only supports sklearn models.")
        tmp_x, tmp_y = working_model.fit_transform(U_hat[:, k:], V_hat[:, k:])
        permed_rs = can_corr(tmp_x, tmp_y)
        cum_r = _wilk_statistic(permed_rs)
        latent_cum_rs.append(cum_r)

    return np.array(latent_cum_rs)


def _get_ndim(model: BaseEstimator):
    try:
        model_params = model.get_params()
        n_comps = model_params["n_components"]
    except:
        raise NotImplementedError("Currently only supports sklearn models.")
    return n_comps


def permutation_inference(
    X: np.ndarray,
    Y: np.ndarray,
    model: Callable or BaseEstimator = None,
    model_kwargs: Dict or Optional = None,
    n_perms: int = 1000,
    fwe_correction: bool = True,
):
    """
    :param X: predictor datasets
    :param Y: target datasets
    :param model: The model can be set or passed as a callable. If callable, then model_kwargs must be set.
    :param model_kwargs:
    :param n_perms:
    :param fwe_correction: Return the cumulated max p-value
    :return
    """
    # TODO: add support for passing model
    # TODO: add support for passing n_jobs
    if X.shape[0] != Y.shape[0]:
        raise ValueError("Y and X do not have same number of rows.")
    if model is None:
        n_comps = min(np.linalg.matrix_rank(X), np.linalg.matrix_rank(Y))
        model_kwargs = {"scale": False}
        working_model = CCA(n_components=n_comps, **model_kwargs)
        model = working_model
    else:
        working_model = model
        if model_kwargs is not None:
            try:
                model.set_params(**model_kwargs)  # sklearn models
            except AttributeError:
                working_model = model(**model_kwargs)
        else:
            model_kwargs = {}
        n_comps = _get_ndim(working_model)

    # Fit the model
    try:
        X_c, Y_c = working_model.fit_transform(X, Y)
    except AttributeError:
        raise AttributeError("Currently only supports sklearn style model.")
    # Get the weights
    try:
        x_weights = working_model.x_weights_
        y_weights = working_model.y_weights_
    except AttributeError:
        # TODO: add support for cca-zoo models
        raise AttributeError("Currently only supports sklearn models.")
    # Expanded canonical variables

    U_hat = X @ np.hstack((x_weights, null_space(x_weights.T)))
    V_hat = Y @ np.hstack((y_weights, null_space(y_weights.T)))
    # Permutation test
    cnt = np.zeros(n_comps)
    seeds = np.random.randint(0, 10000, size=n_perms)
    latent_cum_rs = _calc_cum_rs(U_hat, V_hat, n_comps, model)
    for perm in range(0, n_perms):
        perm_U_hat, perm_V_hat = PermutationMultiVar(
            random_state=seeds[perm], shuffle_target=True, selection=None
        ).fit_transform(U_hat, V_hat)
        permuted_latent_cum_rs = _calc_cum_rs(
            perm_U_hat,
            perm_V_hat,
            n_comps,
            model,
        )
        cnt += (permuted_latent_cum_rs >= latent_cum_rs).astype(int)

    # Compute p-values
    punc = np.minimum(cnt / n_perms, 1)
    if fwe_correction:
        return np.maximum.accumulate(punc)
    else:
        return punc


"""
from concurrent.futures import ThreadPoolExecutor

def permute_and_calc_cum_rs(seed, X_c, Y_c, n_comps):
    perm_X_c, perm_Y_c = PermutationMultiVar(
        random_state=seed, shuffle_target=True, selection=None
    ).fit_transform(X_c, Y_c)
    return _calc_cum_rs(perm_X_c, perm_Y_c, n_comps)

def permutation_inference(X, Y, n_comps=None, n_perms=1000, n_threads=4):
    if X.shape[0] != Y.shape[0]:
        raise ValueError("Y and X do not have same number of rows.")
    if n_comps is None:
        n_comps = min(X.shape[1], Y.shape[1])
    cca = CCA(n_components=n_comps, scale=False)
    cca.fit_transform(X, Y)
    X_c = X @ np.hstack((cca.x_weights_, null_space(cca.x_weights_.T)))
    Y_c = Y @ np.hstack((cca.y_weights_, null_space(cca.y_weights_.T)))
    cnt = np.zeros(n_comps)
    seeds = np.random.randint(0, 10000, size=n_perms)
    latent_cum_rs = _calc_cum_rs(X_c, Y_c, n_comps)
    print(latent_cum_rs)

    with ThreadPoolExecutor(max_workers=n_threads) as executor:
        results = list(executor.map(permute_and_calc_cum_rs, seeds, [X_c]*n_perms, [Y_c]*n_perms, [n_comps]*n_perms))
    
    for permuted_latent_cum_rs in results:
        cnt += (permuted_latent_cum_rs >= latent_cum_rs).astype(int)

    punc = np.minimum(cnt / n_perms, 1)
    return punc
    """
